#!/usr/bin/env python
from datetime import datetime, timedelta
from functools import wraps

import logging
import os
import time
import yaml
import calendar

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as da

from mapies.mapies import MAPIES
from mapies.util.func_tools import *
from mapies.util.cloudsat_config import VARIABLE_DICT
#timeit, get_file_list, time_converter, frequency_int, error_estimation, time_domain_selection, geo_to_rot
from mapies.grids.monarch import RegularGrid
from pathlib import Path

from pyhdf.SD import SD, SDC
from pyhdf.HDF import HDF
import pyhdf.VS




# Uncertainty constants should be able to be passed as number, array or dict

class CLOUDSAT(MAPIES):
    """
    Script for reprocessing of Cloudsat data to a MAPIES format
    """

    def __init__(self, start_date, end_date, **kwargs):
        """
        Inherited init class with new variables for Cloudsat
        """
        super().__init__(start_date, end_date)
        self.datatype="cloudsat"
        self.dest = kwargs.get("dest")
        self.indir = kwargs.get("indir")
        
        if isinstance(self.start_date, str):
            self.start_date = time_converter(self.start_date)

        if isinstance(self.end_date, str):
            self.end_date = time_converter(self.end_date)
        self.read_config(**kwargs)

        self.data = {}


    def read_multiple_files(self):
        # Create a function to read in multiple files in regex and also for the julian days
        # cut down for month using start and end date
        """
        monthly_dates_dict = monthly_dates(self.start_date, self.end_date, output_date_format="%Y%m%d")
        for month, dates in monthly_dates_dict.items():
            print(f"Running analysis for this month {month}")
        """

        dates = pd.date_range(self.start_date, self.end_date).strftime("%Y%j")
        file_patterns = []
        for date in dates:
            print(date)
                    
            filepaths = f'{self.indir}/{date}*'
            print(filepaths)
            file_patterns.append(filepaths)


        filenames = get_file_list(file_patterns)
        print(filenames)
        # Return new dictionary with TAI start
        for filename in filenames:
            da_dict = self.read_hdf(filename)
            da_dict = self.get_geoprof_time(da_dict)
            epoch = tai_base = calendar.timegm((1993, 1, 1, 0, 0, 0)) + da_dict["data_vars"]['TAI_start']["data"]
            print(filename)
            # save each file data in a different time variable each time
            self.data[epoch] = da_dict
        print(self.data)

        

    # I need to connect this to main MAPIES and how we deal with hdf5 files not just netcdf
    @staticmethod
    def read_hdf(filename):
        """
        Open the cloudsat data

        Make this more general for all MAPIES
        """
        da_dict = {
            "coords":{
                "nray":{"dims":"nray", "data":np.arange(0, 36946)},
                "nbin":{"dims":"nbin", "data":np.arange(0, 125)}
            },
            "dims":["nbin", "nray"],
            "data_vars":{}
        }

        # First read SD (scientific datasets)
        sd = SD(filename)
        for dname, info in VARIABLE_DICT.items():

            if dname not in sd.datasets().keys(): 
                continue
            
            sds = sd.select(dname)
            # get (masked) data
            data = np.array(sds[:]).squeeze()
            data = array_calcs(data, info)
            print("Converting array to dask array")
            data = da.from_array(data, chunks=500)

    
            da_dict["data_vars"][dname] = {"data":data, "dims":info["dims"]}
            # Close this dataset
            sds.endaccess()
        # Close file
        sd.end()
        
        # ...now read VDATA...
        hdf = HDF(filename)
        vs = hdf.vstart()
        for vname, info in VARIABLE_DICT.items():
            
            if vname not in [v[0] for v in vs.vdatainfo()]: 
                continue
            
            # attach vdata
            vd = vs.attach(vname)

            # read data
            data = np.array(vd[:]).squeeze()
            data = array_calcs(data, info)
            #data = da.from_array(data, chunks=500)

            if info["dims"] == "scalar":
                da_dict["data_vars"][vname] = {"data":data, "dims":()}
            else:
                da_dict["data_vars"][vname] = {"data":data, "dims":info["dims"]}
                
            vd.detach()
        # clean up
        vs.end()
        # HDF files do not always close cleanly, so close manually
        hdf.close()

        return da_dict
        
    @staticmethod
    def get_geoprof_time(da_dict):
        """
        Read time stamp and return seconds since Epoch
        """

        # read data
        profile_sec = da_dict["data_vars"]['Profile_time']["data"]
        start_tai = da_dict["data_vars"]['TAI_start']["data"]
        # TAI time for each profile
        time_tai = (profile_sec + start_tai)  # seconds since 00:00:00 Jan 1
        tai_base = calendar.timegm((1993, 1, 1, 0, 0, 0))
        # get epoch time as a datetime
        epoch = tai_base + time_tai
        # get array of datetime objects
        data = np.array([datetime.utcfromtimestamp(t) for t in epoch])
        
        da_dict["data_vars"]["time"] = {"data":data, "dims":da_dict["data_vars"]['Profile_time']["dims"]}

        return da_dict



    # This is repeated from VIIRS
    @timeit
    def preprocess_vars(self):
        """
        Preprocessing of the dataset.

        This involves creating a regular 5 by 5 grid 
        """
        #print(self.data)

        r = RegularGrid(centre_lon=0, centre_lat=0, dlon=5, dlat=5, west=-180, south=-90)
        data_dict = {}
        for time, data in self.data.items():
            lon_values = data["data_vars"][self.lon_var]["data"].flatten()
            lat_values = data["data_vars"][self.lat_var]["data"].flatten()
            time_values = data["data_vars"][self.time_var]["data"].flatten()
            for var, info in data["data_vars"].items():
                if var in [self.lon_var, self.lat_var, self.time_var]:
                    continue
                
                # Aggregate the data in the grid and then add to the data dict
                obs = info["data"].flatten()
                # could have a problem here is all arrays are not the same length
                try:
                    lon, lat, obs = r.aggregate(lon_values, lat_values, obs)
                except ValueError:
                    continue
                print(lon)
                print(lat)
                
                # Add the new lon and lat at the end
                # Sum the aggregated variables
                if var in data_dict:
                    a = data_dict[var]
                    b = obs
                    a[np.isnan(a)] = 0  # replace all nan in a with 0
                    b[np.isnan(b)] = 0  # replace all nan in b with 0
                    data_dict[var] = np.add(a, b)
                else:
                    data_dict[var] = obs


        print(var)
        print(data_dict)
        self.lon_values = lon
        self.lat_values = lat
        # Run an aggregate again
        #self.lon_values, self.lat_values, self.obs = r.aggregate(self.lon_values, self.lat_values, data_dict['RO_liq_water_path'])
        self.obs = np.where(data_dict['RO_liq_water_path'] == 0, np.nan, data_dict['RO_liq_water_path'])
        print(self.obs)
        print(self.lon_values)
        


    # Plots
    def plot_2D_obs(self, outdir="./", **kwargs):
        """
        Plotting the observations specific to VIIRS
        """
        
        super().plot_2D_obs()
    

